@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  * 
- * Copyright Ericsson AB 2018-2022. All Rights Reserved.
+ * Copyright Ericsson AB 2018-2023. All Rights Reserved.
  * 
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -110,7 +110,7 @@ typedef struct {
  * Note that not all signal are handled using this functionality!
  */
 
-#define ERTS_SIG_Q_OP_MAX 18
+#define ERTS_SIG_Q_OP_MAX 19
 
 #define ERTS_SIG_Q_OP_EXIT                      0  /* Exit signal due to bif call */
 #define ERTS_SIG_Q_OP_EXIT_LINKED               1  /* Exit signal due to link break*/
@@ -130,7 +130,8 @@ typedef struct {
 #define ERTS_SIG_Q_OP_ALIAS_MSG                 15
 #define ERTS_SIG_Q_OP_RECV_MARK                 16
 #define ERTS_SIG_Q_OP_UNLINK_ACK                17
-#define ERTS_SIG_Q_OP_ADJ_MSGQ                  ERTS_SIG_Q_OP_MAX
+#define ERTS_SIG_Q_OP_ADJ_MSGQ                  18
+#define ERTS_SIG_Q_OP_FLUSH			ERTS_SIG_Q_OP_MAX
 
 #define ERTS_SIG_Q_TYPE_MAX (ERTS_MON_LNK_TYPE_MAX + 10)
 
@@ -204,23 +205,25 @@ extern Eterm erts_old_recv_marker_id;
 #endif
 
 #ifdef ERTS_PROC_SIG_HARD_DEBUG
-#  define ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE(P)       \
-    ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE__((P), "")
+#  define ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE(P, B) \
+    ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE__((P), (B), "")
 #  define ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE(P, QL) \
     ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE__((P), (QL), "")
-#  define ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE__(P, What)   \
-    erts_proc_sig_hdbg_check_in_queue((P), (What), __FILE__, __LINE__)
+#  define ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE__(P, B , What) \
+    erts_proc_sig_hdbg_check_in_queue((P), (B), (What), __FILE__, __LINE__)
 #  define ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE__(P, QL, What)              \
     erts_proc_sig_hdbg_check_priv_queue((P), (QL), (What), __FILE__, __LINE__)
 struct process;
 void erts_proc_sig_hdbg_check_priv_queue(struct process *c_p, int qlock,
                                          char *what, char *file, int line);
-void erts_proc_sig_hdbg_check_in_queue(struct process *c_p, char *what,
-                                       char *file, int line);
+struct ErtsSignalInQueue_;
+void erts_proc_sig_hdbg_check_in_queue(struct process *c_p,
+                                       struct ErtsSignalInQueue_ *buffer,
+                                       char *what, char *file, int line);
 #else
-#  define ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE(P)
+#  define ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE(P, B)
 #  define ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE(P, QL)
-#  define ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE__(P, What)
+#  define ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE__(P, B, What)
 #define ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE__(P, QL, What)
 #endif
 
@@ -245,21 +248,20 @@ void erl_proc_sig_hdbg_chk_recv_marker_block(struct process *c_p);
 void erts_proc_sig_queue_maybe_install_buffers(Process* p, erts_aint32_t state);
 void erts_proc_sig_queue_flush_and_deinstall_buffers(Process* proc);
 void erts_proc_sig_queue_flush_buffers(Process* proc);
-ErtsSignalInQueueBufferArray*
-erts_proc_sig_queue_flush_get_buffers(Process* proc, int *need_unget_buffers);
 void erts_proc_sig_queue_lock(Process* proc);
-ErtsSignalInQueueBufferArray*
+ERTS_GLB_INLINE ErtsSignalInQueueBufferArray*
 erts_proc_sig_queue_get_buffers(Process* p, int *need_unread);
-void erts_proc_sig_queue_unget_buffers(ErtsSignalInQueueBufferArray* buffers,
-                                       int need_unget);
+ERTS_GLB_INLINE void
+erts_proc_sig_queue_unget_buffers(ErtsSignalInQueueBufferArray* buffers,
+                                  int need_unget);
 int erts_proc_sig_queue_try_enqueue_to_buffer(Eterm from,
                                               Process* receiver,
                                               ErtsProcLocks receiver_locks,
                                               ErtsMessage* first,
                                               ErtsMessage** last,
                                               ErtsMessage** last_next,
-                                              Uint len,
-                                              int is_signal);
+                                              Uint len);
+int erts_proc_sig_queue_force_buffers(Process*);
 
 #define ERTS_SIG_Q_OP_BITS      8                      
 #define ERTS_SIG_Q_OP_SHIFT     0
@@ -316,7 +318,9 @@ struct dist_entry_;
 
 #define ERTS_PROC_HAS_INCOMING_SIGNALS(P)                               \
     (!!(erts_atomic32_read_nob(&(P)->state)                             \
-        & (ERTS_PSFLG_SIG_Q|ERTS_PSFLG_SIG_IN_Q)))
+        & (ERTS_PSFLG_SIG_Q                                             \
+           | ERTS_PSFLG_NMSG_SIG_IN_Q                                   \
+           | ERTS_PSFLG_MSG_SIG_IN_Q)))
 
 /*
  * Send operations of currently supported process signals follow...
@@ -862,7 +866,11 @@ erts_proc_sig_send_is_alive_request(Process *c_p, Eterm to,
  * @param[in]     item_ix       Info index array to pass to
  *                              erts_process_info()
  *
- * @param[in]     len           Length of info index array
+ * @param[in]     item_extra    Extra terms array to pass to
+ *                              erts_process_info()
+ *
+ * @param[in]     len           Length of info index array and
+ *                              extra array if such is provided
  *
  * @param[in]     need_msgq_len Non-zero if message queue
  *                              length is needed; otherwise,
@@ -888,6 +896,7 @@ int
 erts_proc_sig_send_process_info_request(Process *c_p,
                                         Eterm to,
                                         int *item_ix,
+                                        Eterm *item_extra,
                                         int len,
                                         int need_msgq_len,
                                         int flags,
@@ -940,7 +949,7 @@ erts_proc_sig_send_sync_suspend(Process *c_p, Eterm to,
  * term returned by 'func'. If the return value of
  * 'func' is not an immediate term, 'func' has to
  * allocate a heap fragment where the result is stored
- * and update the the heap fragment pointer pointer
+ * and update the heap fragment pointer pointer
  * passed as third argument to point to it.
  *
  * If this function returns a reference, 'func' will
@@ -1005,7 +1014,7 @@ erts_proc_sig_send_rpc_request(Process *c_p,
  * term returned by 'func'. If the return value of
  * 'func' is not an immediate term, 'func' has to
  * allocate a heap fragment where the result is stored
- * and update the the heap fragment pointer pointer
+ * and update the heap fragment pointer pointer
  * passed as third argument to point to it.
  *
  * If this function returns a reference, 'func' will
@@ -1132,6 +1141,9 @@ erts_proc_sig_send_move_msgq_off_heap(Eterm to);
  *
  * @param[out]    statep        Pointer to process state after
  *                              signal handling. May not be NULL.
+ *                              The state should recently have
+ *                              been updated before calling
+ *                              this function.
  *
  * @param[in,out] redsp         Pointer to an integer containing
  *                              reductions. On input, the amount
@@ -1185,6 +1197,8 @@ erts_proc_sig_handle_incoming(Process *c_p, erts_aint32_t *statep,
  *                              used by the call. On output, the
  *                              amount of reductions consumed.
  *
+ * @param[in,out] pe_ctxtp      Process exit context pointer.
+ *
  * @return                      Returns a non-zero value, when
  *                              no more signals to handle in the
  *                              middle queue remain. A zero
@@ -1194,8 +1208,7 @@ erts_proc_sig_handle_incoming(Process *c_p, erts_aint32_t *statep,
  */
 int
 erts_proc_sig_handle_exit(Process *c_p, Sint *redsp,
-                          ErtsMonitor **pend_spawn_mon_pp,
-                          Eterm reason);
+                          ErtsProcExitContext *pe_ctxt_p);
 
 /**
  *
@@ -1251,6 +1264,58 @@ erts_proc_sig_receive_helper(Process *c_p, int fcalls,
                              int neg_o_reds, ErtsMessage **msgpp,
                              int *get_outp);
 
+/*
+ * CLEAN_SIGQ - Flush until middle queue is empty, i.e.
+ *              the content of inner+middle queue equals
+ *              the message queue.
+ */
+#define ERTS_PROC_SIG_FLUSH_FLG_CLEAN_SIGQ          (1 << 0)
+/*
+ * FROM_ALL   - Flush signals from all local senders (processes
+ *              and ports).
+ */
+#define ERTS_PROC_SIG_FLUSH_FLG_FROM_ALL            (1 << 1)
+/*
+ * FROM_ID    - Flush signals from process or port identified
+ *              by 'id'.
+ */
+#define ERTS_PROC_SIG_FLUSH_FLG_FROM_ID             (1 << 2)
+
+/*
+ * All erts_proc_sig_init_flush_signals() flags.
+ */
+#define ERTS_PROC_SIG_FLUSH_FLGS                                \
+    (ERTS_PROC_SIG_FLUSH_FLG_CLEAN_SIGQ                         \
+     | ERTS_PROC_SIG_FLUSH_FLG_FROM_ALL                         \
+     | ERTS_PROC_SIG_FLUSH_FLG_FROM_ID)
+
+/**
+ *
+ * @brief Initialize flush of signals from another process or port
+ *
+ * Inserts a flush signal in the outer signal queue of
+ * current process and sets the FS_FLUSHING_SIGS flag in
+ * 'c_p->sig_qs.flags'. When the flush signal has been
+ * handled the FS_FLUSHED_SIGS flag is set as well.
+ *
+ * While the flushing is ongoing the process *should* only
+ * handle incoming signals and not execute Erlang code. When
+ * the functionality that initiated the flush detects that
+ * the flush is done by the FS_FLUSHED_SIGS flag being set,
+ * it should clear both the FS_FLUSHED_SIGS flag and the
+ * FS_FLUSHING_SIGS flag.
+ *
+ * @param[in]   c_p             Pointer to process struct of
+ *                              currently executing process.
+ *              flags           Flags indicating how to flush.
+ *                              (see above).
+ *              from            Identifier of sender to flush
+ *                              signals from in case the
+ *                              FROM_ID flag is set.
+ */
+void
+erts_proc_sig_init_flush_signals(Process *c_p, int flags, Eterm from);
+
 /**
  *
  * @brief Fetch signals from the outer queue
@@ -1288,25 +1353,29 @@ ERTS_GLB_INLINE Sint erts_proc_sig_fetch(Process *p);
 Sint
 erts_proc_sig_privqs_len(Process *c_p);
 
-
 /**
- * @brief Enqueue list of signals on process.
+ * @brief Enqueue a sequence of signals on an in signal queue of
+ *        a process
  *
- * Message queue must be locked on receiving process.
+ * The *only* valid scenarios:
+ * * One or more message signals and no non-message signals.
+ * * One non-message signal followed by one or more message signals
  *
- * @param rp                Receiving process.
- * @param first             First signal in list.
- * @param last              Last signal in list.
- * @param last_next         Pointer to next-pointer to last non-message signal
- *                          or NULL if no non-message signal after 'first'.
- * @param msg_cnt           Number of message signals in list.
- * @param in_state          'state' of rp.
+ * @param rp[in]                 Process to which the in signal queue
+ *                               belong.
+ * @param first[in]              Pointer to the first signal in signal
+ *                               sequence.
+ * @param last[in]               Pointer to the next pointer of the
+ *                               last signal in the sequence. This
+ *                               next pointer should equal NULL.
+ * @param msg_cnt[in]            Number of message signals in seqence
+ * @param in_state[in]           state of rp upon call.
  *
- * @return                  'state' of rp.
+ * @return                       Possibly changed state of rp.
  */
 erts_aint32_t
 erts_enqueue_signals(Process *rp, ErtsMessage *first,
-                     ErtsMessage **last, ErtsMessage **last_next,
+                     ErtsMessage **last,
                      Uint msg_cnt,
                      erts_aint32_t in_state);
 
@@ -1339,10 +1408,14 @@ erts_proc_sig_send_dist_to_alias(Eterm from, Eterm alias,
  */
 ERTS_GLB_INLINE void erts_proc_notify_new_sig(Process* rp, erts_aint32_t state,
                                               erts_aint32_t enable_flag);
+ERTS_GLB_INLINE void erts_proc_notify_new_message(Process *p,
+                                                  ErtsProcLocks locks);
 
-void erts_make_dirty_proc_handled(Eterm pid, erts_aint32_t state,
-                                  erts_aint32_t prio);
-
+void
+erts_ensure_dirty_proc_signals_handled(Process *proc,
+                                       erts_aint32_t state,
+                                       erts_aint32_t prio,
+                                       ErtsProcLocks locks);
 
 typedef struct {
     Uint size;
@@ -1368,12 +1441,16 @@ typedef struct {
  *
  * @param[in]   mip             Pointer to array of
  *                              ErtsMessageInfo structures.
+ *
+ * @param[out]  msgq_lenp       Pointer to integer containing
+ *                              amount of messages.
  */
 Uint erts_proc_sig_prep_msgq_for_inspection(Process *c_p,
                                             Process *rp,
                                             ErtsProcLocks rp_locks,
                                             int info_on_self,
-                                            ErtsMessageInfo *mip);
+                                            ErtsMessageInfo *mip,
+                                            Sint *msgq_lenp);
 
 /**
  *
@@ -1451,6 +1528,26 @@ int
 erts_proc_sig_decode_dist(Process *proc, ErtsProcLocks proc_locks,
                           ErtsMessage *msgp, int force_off_heap);
 
+/**
+ *
+ * @brief Check if a newly scheduled process needs to wait for
+ *        ongoing dirty handling of signals to complete.
+ *
+ * @param[in]   c_p             Pointer to executing process.
+ *
+ * @param[in]   state_in        State of process.
+ *
+ * @return                      Updated state of process if
+ *                              we had to wait; otherwise,
+ *                              state_in.
+ */
+ERTS_GLB_INLINE erts_aint32_t
+erts_proc_sig_check_wait_dirty_handle_signals(Process *c_p,
+                                              erts_aint32_t state_in);
+
+void erts_proc_sig_do_wait_dirty_handle_signals__(Process *c_p);
+
+
 ErtsDistExternal *
 erts_proc_sig_get_external(ErtsMessage *msgp);
 
@@ -1663,8 +1760,12 @@ extern Process *erts_dirty_process_signal_handler_high;
 extern Process *erts_dirty_process_signal_handler_max;
 
 /* Helpers... */
-void erts_proc_sig_fetch__(Process *proc);
-Sint erts_proc_sig_fetch_msgq_len_offs__(Process *proc);
+void erts_proc_sig_fetch__(Process *proc,
+                           ErtsSignalInQueueBufferArray* buffers,
+                           int need_unget_buffers);
+Sint erts_proc_sig_fetch_msgq_len_offs__(Process *proc,
+                                         ErtsSignalInQueueBufferArray* buffers,
+                                         int need_unget_buffers);
 ERTS_GLB_INLINE int erts_msgq_eq_recv_mark_id__(Eterm term1, Eterm term2);
 ERTS_GLB_INLINE void erts_msgq_recv_marker_set_save__(Process *c_p,
 				 ErtsRecvMarkerBlock *blkp,
@@ -1674,7 +1775,7 @@ Eterm erts_msgq_recv_marker_create_insert(Process *c_p, Eterm id);
 void erts_msgq_recv_marker_create_insert_set_save(Process *c_p, Eterm id);
 ErtsMessage **erts_msgq_pass_recv_markers(Process *c_p,
 					  ErtsMessage **markpp);
-void erts_msgq_remove_leading_recv_markers(Process *c_p);
+void erts_msgq_remove_leading_recv_markers_set_save_first(Process *c_p);
 
 #define ERTS_RECV_MARKER_IX__(BLKP, MRKP) \
     ((int) ((MRKP) - &(BLKP)->marker[0]))
@@ -1709,6 +1810,48 @@ erts_proc_sig_new_unlink_id(ErtsPTabElementCommon *sender)
     return id;
 }
 
+ERTS_GLB_INLINE ErtsSignalInQueueBufferArray*
+erts_proc_sig_queue_get_buffers(Process* p, int *need_unread)
+{
+    ErtsThrPrgrDelayHandle dhndl = erts_thr_progress_unmanaged_delay();
+    ErtsSignalInQueueBufferArray* buffers =
+        (ErtsSignalInQueueBufferArray*)erts_atomic_read_acqb(&p->sig_inq_buffers);
+    *need_unread = 0;
+    if (ERTS_THR_PRGR_DHANDLE_MANAGED == dhndl) {
+        erts_thr_progress_unmanaged_continue(dhndl);
+        return buffers;
+    }
+    if (buffers == NULL) {
+        erts_thr_progress_unmanaged_continue(dhndl);
+        return NULL;
+    }
+    erts_refc_inc(&buffers->dirty_refc, 2);
+    erts_thr_progress_unmanaged_continue(dhndl);
+    *need_unread = 1;
+    return buffers;
+}
+
+ERTS_GLB_INLINE void
+erts_proc_sig_queue_unget_buffers(ErtsSignalInQueueBufferArray* buffers,
+                                  int need_unget)
+{
+    if (!need_unget) {
+        return;
+    } else {
+        int i;
+        erts_aint_t refc = erts_refc_dectest(&buffers->dirty_refc, 0);
+        if (refc != 0) {
+            return;
+        }
+        ASSERT(!buffers->alive);
+        for (i = 0; i < ERTS_PROC_SIG_INQ_BUFFERED_NR_OF_BUFFERS; i++) {
+            ASSERT(!buffers->slots[i].b.alive);
+            erts_mtx_destroy(&buffers->slots[i].b.lock);
+        }
+        erts_free(ERTS_ALC_T_SIGQ_BUFFERS, buffers);
+    }
+}
+
 ERTS_GLB_INLINE Sint
 erts_proc_sig_fetch(Process *proc)
 {
@@ -1716,40 +1859,41 @@ erts_proc_sig_fetch(Process *proc)
     ErtsSignal *sig;
     ErtsSignalInQueueBufferArray* buffers;
     int need_unget_buffers;
-    ERTS_LC_ASSERT(ERTS_PROC_IS_EXITING(proc)
-                   || ((erts_proc_lc_my_proc_locks(proc)
-                        & (ERTS_PROC_LOCK_MAIN
-                           | ERTS_PROC_LOCK_MSGQ))
-                       == (ERTS_PROC_LOCK_MAIN
-                           | ERTS_PROC_LOCK_MSGQ)));
+    ERTS_LC_ASSERT((erts_proc_lc_my_proc_locks(proc)
+                    & (ERTS_PROC_LOCK_MAIN
+                       | ERTS_PROC_LOCK_MSGQ))
+                   == (ERTS_PROC_LOCK_MAIN
+                       | ERTS_PROC_LOCK_MSGQ));
 
-    ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE(proc);
+    ASSERT(!(proc->sig_qs.flags & FS_FLUSHING_SIGS)
+           || ERTS_PROC_IS_EXITING(proc)
+           || ERTS_IS_CRASH_DUMPING);
+
+    ASSERT(!(proc->sig_qs.flags & FS_HANDLING_SIGS));
+
+    ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE(proc, &proc->sig_inq);
     ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE(proc, !0);
 
-    buffers = erts_proc_sig_queue_flush_get_buffers(proc,
-                                                    &need_unget_buffers);
+    proc->sig_qs.flags &= ~FS_NON_FETCH_CNT_MASK;
+
+    buffers = erts_proc_sig_queue_get_buffers(proc, &need_unget_buffers);
 
     sig = (ErtsSignal *) proc->sig_inq.first;
-    if (sig) {
-        if (ERTS_LIKELY(sig->common.tag != ERTS_PROC_SIG_MSGQ_LEN_OFFS_MARK))
-            erts_proc_sig_fetch__(proc);
-        else
-            res = erts_proc_sig_fetch_msgq_len_offs__(proc);
+
+    if (!sig) {
+        if (buffers)
+            goto fetch;
     }
-    if (buffers) {
-        Uint32 state = erts_atomic32_read_acqb(&proc->state);
-        if (!(ERTS_PSFLG_SIG_IN_Q & state) &&
-            erts_atomic64_read_nob(&buffers->nonmsg_slots)) {
-            /* We may have raced with a thread inserting into a buffer
-             * when resetting the flag ERTS_PSFLG_SIG_IN_Q in one of
-             * the fetch functions above so we have to make sure that
-             * it is set when there is a nonmsg signal in the buffers. */
-            erts_atomic32_read_bor_nob(&proc->state,
-                                        ERTS_PSFLG_SIG_IN_Q |
-                                        ERTS_PSFLG_ACTIVE);
-        }
-        erts_proc_sig_queue_unget_buffers(buffers, need_unget_buffers);
+    else if (ERTS_UNLIKELY(sig->common.tag
+                           == ERTS_PROC_SIG_MSGQ_LEN_OFFS_MARK)) {
+        res = erts_proc_sig_fetch_msgq_len_offs__(proc, buffers,
+                                                  need_unget_buffers);
     }
+    else {
+    fetch:
+        erts_proc_sig_fetch__(proc, buffers, need_unget_buffers);
+    }
+
     res += proc->sig_qs.len;
 
     ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE(proc, !0);
@@ -1774,24 +1918,32 @@ ERTS_GLB_INLINE void
 erts_proc_notify_new_sig(Process* rp, erts_aint32_t state,
                          erts_aint32_t enable_flag)
 {
-    if (~(state & (ERTS_PSFLG_EXITING
-                   | ERTS_PSFLG_ACTIVE_SYS
-                   | ERTS_PSFLG_SIG_IN_Q))
-        | (~state & enable_flag)) {
+    if ((!(state & (ERTS_PSFLG_EXITING
+                    | ERTS_PSFLG_ACTIVE_SYS))) | (~state & enable_flag)) {
         /* Schedule process... */
         state = erts_proc_sys_schedule(rp, state, enable_flag);
     }
 
-    if (state & ERTS_PSFLG_DIRTY_RUNNING) {
-        /*
-         * We ignore ERTS_PSFLG_DIRTY_RUNNING_SYS. For
-         * more info see erts_execute_dirty_system_task()
-         * in erl_process.c.
-         */
-        erts_make_dirty_proc_handled(rp->common.id, state, -1);
+    if (ERTS_PROC_IN_DIRTY_STATE(state)) {
+        erts_ensure_dirty_proc_signals_handled(rp, state, -1, 0);
     }
 }
 
+
+
+ERTS_GLB_INLINE void
+erts_proc_notify_new_message(Process *p, ErtsProcLocks locks)
+{
+    /* No barrier needed, due to msg lock */
+    erts_aint32_t state = erts_atomic32_read_nob(&p->state);
+    if (!(state & ERTS_PSFLG_ACTIVE))
+	erts_schedule_process(p, state, locks);
+    if (ERTS_PROC_NEED_DIRTY_SIG_HANDLING(state)) {
+        erts_ensure_dirty_proc_signals_handled(p, state, -1, locks);
+    }
+}
+
+
 #undef ERTS_PROC_SIG_RECV_MARK_CLEAR_PENDING_SET_SAVE__
 #define ERTS_PROC_SIG_RECV_MARK_CLEAR_PENDING_SET_SAVE__(BLKP) 		\
     do {								\
@@ -1899,6 +2051,7 @@ erts_msgq_recv_marker_clear(Process *c_p, Eterm id)
 {
     ErtsRecvMarkerBlock *blkp = c_p->sig_qs.recv_mrk_blk;
     int ix;
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
 
     if (!is_small(id) && !is_big(id) && !is_internal_ref(id))
 	return;
@@ -1926,6 +2079,7 @@ erts_msgq_recv_marker_clear(Process *c_p, Eterm id)
 ERTS_GLB_INLINE Eterm
 erts_msgq_recv_marker_insert(Process *c_p)
 {
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
     erts_proc_sig_queue_lock(c_p);
     erts_proc_sig_fetch(c_p);
     erts_proc_unlock(c_p, ERTS_PROC_LOCK_MSGQ);
@@ -1939,6 +2093,7 @@ ERTS_GLB_INLINE void erts_msgq_recv_marker_bind(Process *c_p,
 						Eterm insert_id,
 						Eterm bind_id)
 {
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
 #ifdef ERTS_SUPPORT_OLD_RECV_MARK_INSTRS
     ASSERT(bind_id != erts_old_recv_marker_id);
 #endif
@@ -1968,6 +2123,7 @@ ERTS_GLB_INLINE void erts_msgq_recv_marker_bind(Process *c_p,
 ERTS_GLB_INLINE void
 erts_msgq_recv_marker_insert_bind(Process *c_p, Eterm id)
 {
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
     if (is_internal_ref(id)) {
 #ifdef ERTS_SUPPORT_OLD_RECV_MARK_INSTRS
 	ErtsRecvMarkerBlock *blkp = c_p->sig_qs.recv_mrk_blk;
@@ -1987,6 +2143,7 @@ erts_msgq_recv_marker_insert_bind(Process *c_p, Eterm id)
 ERTS_GLB_INLINE void
 erts_msgq_recv_marker_set_save(Process *c_p, Eterm id)
 {
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
     if (is_internal_ref(id)) {
 	ErtsRecvMarkerBlock *blkp = c_p->sig_qs.recv_mrk_blk;
 
@@ -2007,6 +2164,7 @@ erts_msgq_recv_marker_set_save(Process *c_p, Eterm id)
 ERTS_GLB_INLINE ErtsMessage *
 erts_msgq_peek_msg(Process *c_p)
 {
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
     ASSERT(!(*c_p->sig_qs.save) || ERTS_SIG_IS_MSG(*c_p->sig_qs.save));
     return *c_p->sig_qs.save;
 }
@@ -2015,6 +2173,7 @@ ERTS_GLB_INLINE void
 erts_msgq_unlink_msg(Process *c_p, ErtsMessage *msgp)
 {
     ErtsMessage *sigp = msgp->next;
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
     ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE__(c_p, 0, "before");
     *c_p->sig_qs.save = sigp;
     c_p->sig_qs.len--;
@@ -2033,6 +2192,7 @@ ERTS_GLB_INLINE void
 erts_msgq_set_save_first(Process *c_p)
 {
     ErtsRecvMarkerBlock *blkp = c_p->sig_qs.recv_mrk_blk;
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
     if (blkp) {
 	ERTS_PROC_SIG_RECV_MARK_CLEAR_PENDING_SET_SAVE__(blkp);
 #ifdef ERTS_SUPPORT_OLD_RECV_MARK_INSTRS
@@ -2046,14 +2206,16 @@ erts_msgq_set_save_first(Process *c_p)
      * anymore...
      */
     if (c_p->sig_qs.first && ERTS_SIG_IS_RECV_MARKER(c_p->sig_qs.first))
-	erts_msgq_remove_leading_recv_markers(c_p);
-    c_p->sig_qs.save = &c_p->sig_qs.first;
+	erts_msgq_remove_leading_recv_markers_set_save_first(c_p);
+    else
+        c_p->sig_qs.save = &c_p->sig_qs.first;
 }
 
 ERTS_GLB_INLINE void
 erts_msgq_unlink_msg_set_save_first(Process *c_p, ErtsMessage *msgp)
 {
     ErtsMessage *sigp = msgp->next;
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
     ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE__(c_p, 0, "before");
     *c_p->sig_qs.save = sigp;
     c_p->sig_qs.len--;
@@ -2070,6 +2232,7 @@ erts_msgq_set_save_next(Process *c_p)
 {
     ErtsMessage *sigp = (*c_p->sig_qs.save)->next;
     ErtsMessage **sigpp = &(*c_p->sig_qs.save)->next;
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
     ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE(c_p, 0);
     if (sigp && ERTS_SIG_IS_RECV_MARKER(sigp))
         sigpp = erts_msgq_pass_recv_markers(c_p, sigpp);
@@ -2081,6 +2244,7 @@ ERTS_GLB_INLINE void
 erts_msgq_set_save_end(Process *c_p)
 {
     /* Set save pointer to end of message queue... */
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
 
     erts_proc_sig_queue_lock(c_p);
     erts_proc_sig_fetch(c_p);
@@ -2100,6 +2264,23 @@ erts_msgq_set_save_end(Process *c_p)
 #undef ERTS_PROC_SIG_RECV_MARK_CLEAR_PENDING_SET_SAVE__
 #undef ERTS_PROC_SIG_RECV_MARK_CLEAR_OLD_MARK__
 
+ERTS_GLB_INLINE erts_aint32_t
+erts_proc_sig_check_wait_dirty_handle_signals(Process *c_p,
+                                              erts_aint32_t state_in)
+{
+    erts_aint32_t state = state_in;
+    ASSERT(erts_get_scheduler_data()->type == ERTS_SCHED_NORMAL);
+    ERTS_LC_ASSERT(ERTS_PROC_LOCK_MAIN == erts_proc_lc_my_proc_locks(c_p));
+
+    if (c_p->sig_qs.flags & FS_HANDLING_SIGS) {
+        erts_proc_sig_do_wait_dirty_handle_signals__(c_p);
+        state = erts_atomic32_read_mb(&c_p->state);
+    }
+    ERTS_LC_ASSERT(ERTS_PROC_LOCK_MAIN == erts_proc_lc_my_proc_locks(c_p));
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
+    return state;
+}
+
 #endif /* ERTS_GLB_INLINE_INCL_FUNC_DEF */
 
 #endif /* ERTS_PROC_SIG_QUEUE_H__ */
